;; Generated package description from expreg.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "expreg" "1.4.1.0.20260619.7" "Simple expand region" '((emacs "29.1")) :commit "d3ac3703e3e0aa66dc1ac6f8110b1862206beb20" :authors '(("Yuan Fu" . "casouri@gmail.com")) :maintainer '("Yuan Fu" . "casouri@gmail.com") :keywords '("text" "editing") :url "https://github.com/casouri/expreg")
